<?php

namespace Drupal\template_entities\Plugin\TemplatePlugin;

use Drupal;
use Drupal\Component\Plugin\PluginBase;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Entity\EntityChangedInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\Entity\RevisionLogInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginWithFormsTrait;
use Drupal\Core\Routing\Router;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\template_entities\Entity\Template;
use Drupal\template_entities\Entity\TemplateType;
use Drupal\template_entities\Plugin\TemplatePluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Routing\RouteCollection;

/**
 * Base class for Template plugin plugins.
 *
 * Plugins needing configuration should override:
 *  - defaultConfiguration()
 *  - templateForm($form, $form_state)
 *  - templateValidate($form, $form_state)
 *  - templateSubmit($form, $form_state)
 *
 * @TemplatePlugin(
 *   id = "canonical_entities",
 *   deriver =
 *   "\Drupal\template_entities\Plugin\Deriver\EntityTemplatePluginDeriver"
 * )
 */
class TemplatePluginBase extends PluginBase implements TemplatePluginInterface, ContainerFactoryPluginInterface {

  use PluginWithFormsTrait;
  use StringTranslationTrait;

  /**
   * The plugin entity type.
   *
   * @var \Drupal\Core\Entity\EntityTypeInterface
   */
  protected EntityTypeInterface $entityType;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The router.
   *
   * @var \Drupal\Core\Routing\Router
   */
  protected Router $router;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected EntityTypeBundleInfoInterface $entityTypeBundleInfo;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeInterface $entity_type, EntityTypeManagerInterface $entity_type_manager, Router $router, EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityType = $entity_type;
    $this->entityTypeManager = $entity_type_manager;
    $this->router = $router;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;


    // Set configuration with defaults.
    $this->setConfiguration($configuration);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager')
        ->getDefinition($plugin_definition['entity_type_id']),
      $container->get('entity_type.manager'),
      $container->get('router.no_access_checks'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function duplicateEntity(EntityInterface $entity, Template $template) {
    // Simple use.
    $duplicate = $entity->createDuplicate();

    $duplicate->template_original = $entity;
    $duplicate->template = $template;

    $request_time = Drupal::time()
      ->getRequestTime();

    if ($entity instanceof EntityChangedInterface) {
      /** @var EntityChangedInterface $duplicate */
      $duplicate->setChangedTime($request_time);
    }

    if ($entity instanceof RevisionLogInterface) {
      /** @var RevisionLogInterface $duplicate */
      $duplicate->setRevisionLogMessage($this->t('Created from template.'));
    }

    return $duplicate;
  }

  /**
   * {@inheritdoc}
   */
  public function getCollectionLinkTemplate() {
    return $this->entityType->hasLinkTemplate('collection') ? $this->entityType->getLinkTemplate('collection') : Url::fromRoute('system.admin_content')
      ->toString();
  }

  /**
   * {@inheritdoc}
   */
  public function getCollectionRoute(RouteCollection $route_collection) {
    $collection_route_collection = new RouteCollection();

    if ($route = $route_collection->get("entity.{$this->getEntityTypeId()}.collection")) {
      $collection_route_collection->add("entity.{$this->getEntityTypeId()}.collection", $route);
    }
    elseif ($route = $route_collection->get("{$this->getEntityTypeId()}.collection")) {
      $collection_route_collection->add("{$this->getEntityTypeId()}.collection", $route);
    }
    else {
      $collection_route_collection->add('system.admin_content', $route_collection->get('system.admin_content'));
    }

    return $collection_route_collection;
  }

  /**
   * Returns the entity type ID.
   *
   * @return string
   *   The entity type ID.
   */
  protected function getEntityTypeId() {
    return $this->entityType->id();
  }

  /**
   * Returns the entity type.
   *
   * @return EntityTypeInterface
   *   The entity type ID.
   */
  public function getEntityType() {
    return $this->entityType;
  }

  /**
   * Get an array of bundles that the plugin supports.
   *
   * @return array
   *   Array of bundles that this plugin supports.
   */
  public function getBundleOptions() {
    return $this->entityTypeBundleInfo->getBundleInfo($this->entityType->id());
  }

  /**
   * {@inheritdoc}
   */
  public function selectAlter(SelectInterface $query, TemplateType $template_type) {
  }

  /**
   * {@inheritdoc}
   */
  public function entityQueryAlter(SelectInterface $query, TemplateType $template_type) {
    // Add default entity query joins.

    // This method not called for all revision queries so safe to use
    // getBaseTable().
    $base_table = $this->entityType->getBaseTable();
    $id_field = $this->entityType->getKey('id');
    $this->selectAddTemplateFilter($query, $template_type, $base_table, $id_field);

    $query->addTag('template_entities_filtered');
  }


  /**
   * Add template filter.
   *
   * @param \Drupal\Core\Database\Query\SelectInterface $query
   * @param \Drupal\template_entities\Entity\TemplateType $template_type
   * @param string|null $base_table
   * @param string|null $id_field
   */
  protected function selectAddTemplateFilter(SelectInterface $query, TemplateType $template_type, string $base_table = NULL, string $id_field = NULL) {
    $base_table = $base_table ?: $this->entityType->getBaseTable();

    // Check for alias.
    if (!isset($query->getTables()[$base_table])) {
      foreach ($query->getTables() as $table) {
        if ($table['table'] === $base_table) {
          $base_table = $table['alias'];
          break;
        }
      }
    }

    $id_field = $id_field ?: $this->entityType->getKey('id');

    // Left join to the template entity reference field table to link entities
    // with references from template entities that have for template types
    // (template bundles) that apply to the entity type.
    $template_table_alias = $query->leftJoin('template__template_entity_id', 'template__template_entity_id', "(%alias.template_entity_id_target_id = $base_table.$id_field AND %alias.bundle = :type)", [':type' => $template_type->id()]);

    // Exclude all rows where there is a valid join to the template entity
    // reference field table.
    $query->isNull($template_table_alias . '.template_entity_id_target_id');
  }

  /**
   * @inheritDoc
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    // Add plugin-specific settings for this block type.
    $form += $this->templateForm($form, $form_state);
    return $form;
  }

  /**
   * Plugin specific configuration form.
   *
   * @param array $form
   *   The form definition array for the full template configuration form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  protected function templateForm($form, FormStateInterface $form_state) {
    return [];
  }

  /**
   * @inheritDoc
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->templateValidate($form, $form_state);
  }

  /**
   * Plugin specific configuration form validation.
   *
   * @param array $form
   *   The form definition array for the full template configuration form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  protected function templateValidate($form, FormStateInterface $form_state) {
  }

  /**
   * @inheritDoc
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getErrors()) {
      $this->templateSubmit($form, $form_state);
    }
  }

  /**
   * Plugin specific configuration form submission.
   *
   * @param array $form
   *   The form definition array for the full template configuration form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  protected function templateSubmit($form, FormStateInterface $form_state) {
  }

  /**
   * @inheritDoc
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * @inheritDoc
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = NestedArray::mergeDeep(
      $this->baseConfigurationDefaults(),
      $this->defaultConfiguration(),
      $configuration
    );
  }

  /**
   * @inheritDoc
   */
  public function defaultConfiguration() {
    return [];
  }

  /**
   * Returns generic default configuration for template plugins.
   *
   * @return array
   *   An associative array with the default configuration.
   */
  protected function baseConfigurationDefaults() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function duplicateEntityInsert(EntityInterface $entity) {}
}
